#include <SPI.h>
#include <MFRC522.h>
#include <LiquidCrystal_I2C.h>
#include <Wire.h>
#include "thingProperties.h"

#define RST_PIN         0
#define SS_PIN          5

#define LED_RGB_RED_PIN   14
#define LED_RGB_GREEN_PIN 12
#define LED_RGB_BLUE_PIN  25

#define BUZZER_PIN        27

MFRC522 mfrc522(SS_PIN, RST_PIN); 

LiquidCrystal_I2C lcd(0x27, 20, 4);

bool cardPresent = false;

struct User {
    String uid;
    String name;
    bool isInside;
};

User users[] = {
    {"84850FC9", "Alice", false},
    {"849E19C9", "Bob", false},
    {"5DEA616F", "Charlie", false}
};

const int numUsers = sizeof(users) / sizeof(users[0]);

User* getUserByUID(const String& uid) {
    for (int i = 0; i < numUsers; i++) {
        if (users[i].uid == uid) {
            return &users[i];
        }
    }
    return nullptr;
}

String getUIDString(MFRC522::Uid* uid) {
    String uidString = "";
    for (byte i = 0; i < uid->size; i++) {
        if (uid->uidByte[i] < 0x10) uidString += "0";
        uidString += String(uid->uidByte[i], HEX);
    }
    uidString.toUpperCase();
    return uidString;
}

bool isValidUID(const String& uid) {
    return getUserByUID(uid) != nullptr;
}

void setup() {
    Serial.begin(9600);
    SPI.begin();
    mfrc522.PCD_Init();
  
    pinMode(LED_RGB_RED_PIN, OUTPUT);
    pinMode(LED_RGB_GREEN_PIN, OUTPUT);
    pinMode(LED_RGB_BLUE_PIN, OUTPUT);
    pinMode(BUZZER_PIN, OUTPUT);
  
    lcd.init();
    lcd.backlight();
    lcd.setCursor(4, 1);
    lcd.print("WAITING FOR");
    lcd.setCursor(3, 2);
    lcd.print("CONNECTION   "); 

    Serial.println(F("Read personal data on a MIFARE PICC:"));
    initProperties();

    ArduinoCloud.begin(ArduinoIoTPreferredConnection);

    int dotCount = 0;
    while (!ArduinoCloud.connected()) {
        lcd.setCursor(14, 2);
        for (int i = 0; i < 3; i++) {
            if (i < dotCount) {
                lcd.print(".");
            } else {
                lcd.print(" ");
            }
        }

        dotCount = (dotCount + 1) % 4;

        digitalWrite(LED_RGB_BLUE_PIN, HIGH);
        digitalWrite(LED_RGB_GREEN_PIN, HIGH);
        digitalWrite(LED_RGB_RED_PIN, HIGH);
        delay(50);
        digitalWrite(LED_RGB_BLUE_PIN, LOW);
        digitalWrite(LED_RGB_GREEN_PIN, LOW);
        digitalWrite(LED_RGB_RED_PIN, LOW);
        delay(50);
        digitalWrite(LED_RGB_BLUE_PIN, HIGH);
        digitalWrite(LED_RGB_GREEN_PIN, HIGH);
        digitalWrite(LED_RGB_RED_PIN, HIGH);
        delay(50);
        digitalWrite(LED_RGB_BLUE_PIN, LOW);
        digitalWrite(LED_RGB_GREEN_PIN, LOW);
        digitalWrite(LED_RGB_RED_PIN, LOW);
        delay(700);
        digitalWrite(LED_RGB_BLUE_PIN, HIGH);
        digitalWrite(LED_RGB_GREEN_PIN, HIGH);
        digitalWrite(LED_RGB_RED_PIN, HIGH);
        delay(50);
        digitalWrite(LED_RGB_BLUE_PIN, LOW);
        digitalWrite(LED_RGB_GREEN_PIN, LOW);
        digitalWrite(LED_RGB_RED_PIN, LOW);
        delay(50);
        digitalWrite(LED_RGB_BLUE_PIN, HIGH);
        digitalWrite(LED_RGB_GREEN_PIN, HIGH);
        digitalWrite(LED_RGB_RED_PIN, HIGH);
        delay(50);
        digitalWrite(LED_RGB_BLUE_PIN, LOW);
        digitalWrite(LED_RGB_GREEN_PIN, LOW);
        digitalWrite(LED_RGB_RED_PIN, LOW);
        delay(600);
        ArduinoCloud.update();
    }

    tone(BUZZER_PIN, 5000);
    digitalWrite(LED_RGB_BLUE_PIN, HIGH);
    digitalWrite(LED_RGB_GREEN_PIN, HIGH);
    digitalWrite(LED_RGB_RED_PIN, HIGH);
    lcd.clear();
    lcd.setCursor(5, 1);
    lcd.print("CONNECTION");
    lcd.setCursor(4, 2);
    lcd.print("ESTABLISHED!");
    delay(300);
    noTone(BUZZER_PIN);
    delay(1500);
    digitalWrite(LED_RGB_BLUE_PIN, LOW);
    digitalWrite(LED_RGB_GREEN_PIN, LOW);
    digitalWrite(LED_RGB_RED_PIN, LOW);
    lcd.clear();
    delay(1500);
    lcd.setCursor(6, 1);
    lcd.print("Welcome!");
    lcd.setCursor(3, 2);
    lcd.print("Put your card");
    setDebugMessageLevel(2);
    ArduinoCloud.printDebugInfo();
}

void loop() {
    ArduinoCloud.update();
    digitalWrite(LED_RGB_RED_PIN, LOW);
    digitalWrite(LED_RGB_GREEN_PIN, LOW);

    MFRC522::MIFARE_Key key;
    for (byte i = 0; i < 6; i++) key.keyByte[i] = 0xFF;

    if (!mfrc522.PICC_IsNewCardPresent()) {
        if (cardPresent) {
            cardPresent = false;
            lcd.clear();
            lcd.setCursor(6, 1);
            lcd.print("Welcome!");
            lcd.setCursor(3, 2);
            lcd.print("Put your card");
        }
        return;
    }

    if (!cardPresent) {
        cardPresent = true;
        lcd.clear();
    }

    if (!mfrc522.PICC_ReadCardSerial()) {
        return;
    }

    String uidString = getUIDString(&mfrc522.uid);
    Serial.println(F("**Card Detected:**"));
    Serial.println(uidString);

    if (isValidUID(uidString)) {
        User* user = getUserByUID(uidString);

        if (user != nullptr) {
            tone(BUZZER_PIN, 2000); 
            digitalWrite(LED_RGB_GREEN_PIN, HIGH); 
            delay(100);
            noTone(BUZZER_PIN); 

            if (user->isInside) {
                String cloudMessage = user->name + " left";
                log_access = cloudMessage;
                lcd.setCursor(3, 1);
                lcd.print("Goodbye,");
            } else {
                String cloudMessage = user->name + " entered";
                log_access = cloudMessage;
                lcd.setCursor(2, 1);
                lcd.print("Have a nice day,");
            }
            user->isInside = !user->isInside;

            int nameLength = user->name.length();
            int start_col = (20 - nameLength) / 2;
            lcd.setCursor(start_col, 2);
            lcd.print(user->name);
            lcd.setCursor(start_col + nameLength, 2);
            lcd.print("!");

            delay(2000); 
            lcd.clear();  
            digitalWrite(LED_RGB_GREEN_PIN, LOW); 
        }
    } else {
        String cloudRejectMessage = "Someone tried to access the gate";
        log_access = cloudRejectMessage;
        lcd.setCursor(3, 1);
        lcd.print("Access Denied !");
        tone(BUZZER_PIN, 1500);
        digitalWrite(LED_RGB_RED_PIN, HIGH);
        delay(100);
        noTone(BUZZER_PIN);
        digitalWrite(LED_RGB_RED_PIN, LOW);

        delay(50);

        tone(BUZZER_PIN, 1500);
        digitalWrite(LED_RGB_RED_PIN, HIGH);
        delay(100);
        noTone(BUZZER_PIN);
        digitalWrite(LED_RGB_RED_PIN, LOW);

        delay(50);

        tone(BUZZER_PIN, 1500);
        digitalWrite(LED_RGB_RED_PIN, HIGH);
        delay(400);
        noTone(BUZZER_PIN);
        digitalWrite(LED_RGB_RED_PIN, LOW);

        delay(2000);
    }

    mfrc522.PICC_HaltA();
    mfrc522.PCD_StopCrypto1();
}

void onOpenGateChange()  {
    lcd.clear();
    tone(BUZZER_PIN, 2000); 
    lcd.setCursor(8, 1);
    digitalWrite(LED_RGB_GREEN_PIN, HIGH); 
    lcd.print("OPEN");
    delay(100);
    noTone(BUZZER_PIN); 
    delay(1500);
    digitalWrite(LED_RGB_GREEN_PIN, LOW); 
    lcd.clear();
    log_access = "Open by Admin";
    delay(300);

    lcd.setCursor(6, 1);
    lcd.print("Welcome!");
    lcd.setCursor(3, 2);
    lcd.print("Put your card");

    loop();
}
